/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * cairo-compmgr
 * Copyright (C) Nicolas Bruguier 2007 <gandalfn@club-internet.fr>
 * 
 * cairo-compmgr is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * cairo-compmgr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with cairo-compmgr.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */
#include <stdio.h>
#include <math.h>

#include "ccm-drawable.h"
#include "ccm-window.h"
#include "ccm-display.h"
#include "ccm-screen.h"
#include "ccm-extension-loader.h"
#include "ccm-magnifier.h"
#include "ccm.h"

enum
{
	CCM_MAGNIFIER_ZOOM_LEVEL,
	CCM_MAGNIFIER_X,
	CCM_MAGNIFIER_Y,
	CCM_MAGNIFIER_HEIGHT,
	CCM_MAGNIFIER_WIDTH,
	CCM_MAGNIFIER_OPTION_N
};

static gchar* CCMMagnifierOptions[CCM_MAGNIFIER_OPTION_N] = {
	"zoom-level",
	"x",
	"y",
	"height",
	"width"
};

static void ccm_magnifier_screen_iface_init(CCMScreenPluginClass* iface);
static void ccm_magnifier_window_iface_init(CCMWindowPluginClass* iface);

CCM_DEFINE_PLUGIN (CCMMagnifier, ccm_magnifier, CCM_TYPE_PLUGIN, 
				   CCM_IMPLEMENT_INTERFACE(ccm_magnifier,
										   CCM_TYPE_WINDOW_PLUGIN,
										   ccm_magnifier_window_iface_init);
				   CCM_IMPLEMENT_INTERFACE(ccm_magnifier,
										   CCM_TYPE_SCREEN_PLUGIN,
										   ccm_magnifier_screen_iface_init))

struct _CCMMagnifierPrivate
{	
	CCMWindow*     window;
	
	cairo_t*       ctx;
	
	CCMConfig*     options[CCM_MAGNIFIER_OPTION_N];
};

#define CCM_MAGNIFIER_GET_PRIVATE(o)  \
   ((CCMMagnifierPrivate*)G_TYPE_INSTANCE_GET_PRIVATE ((o), CCM_TYPE_MAGNIFIER, CCMMagnifierClass))

enum
{
	CCM_SCREEN_BACKEND,
	CCM_SCREEN_PIXMAP,
	CCM_SCREEN_USE_BUFFERED,
	CCM_SCREEN_PLUGINS,
	CCM_SCREEN_REFRESH_RATE,
	CCM_SCREEN_SYNC_WITH_VBLANK,
	CCM_SCREEN_OPTION_N
};

static gchar* CCMScreenOptions[CCM_SCREEN_OPTION_N] = {
	"backend",
	"native_pixmap_bind",
	"use_buffered_pixmap",
	"plugins",
	"refresh_rate",
	"sync_with_vblank"
};

struct _CCMScreenPrivate
{
	CCMDisplay* 		display;
	
	cairo_t*			ctx;
	
	CCMWindow* 			root;
	CCMWindow* 			cow;
	CCMWindow*          fullscreen;
	
	CCMRegion*			damaged;
	
	GList*				windows;
	
	GTimer*				vblank;
	guint				id_paint;
	
	CCMExtensionLoader* plugin_loader;
	CCMScreenPlugin*	plugin;

	GSList*				animations;
	CCMConfig*			options[CCM_SCREEN_OPTION_N];
};

static void
ccm_magnifier_init (CCMMagnifier *self)
{
	self->priv = CCM_MAGNIFIER_GET_PRIVATE(self);
	self->priv->window = NULL;
	self->priv->ctx = NULL;
}

static void
ccm_magnifier_finalize (GObject *object)
{
	CCMMagnifier* self = CCM_MAGNIFIER(object);
	
	G_OBJECT_CLASS (ccm_magnifier_parent_class)->finalize (object);
}

static void
ccm_magnifier_class_init (CCMMagnifierClass *klass)
{
	GObjectClass* object_class = G_OBJECT_CLASS (klass);
	
	g_type_class_add_private (klass, sizeof (CCMMagnifierPrivate));

	object_class->finalize = ccm_magnifier_finalize;
}

static void
ccm_magnifier_screen_load_options(CCMScreenPlugin* plugin, CCMScreen* screen)
{
	CCMMagnifier* self = CCM_MAGNIFIER(plugin);
	gint cpt;
	
	for (cpt = 0; cpt < CCM_MAGNIFIER_OPTION_N; cpt++)
	{
		self->priv->options[cpt] = ccm_config_new(screen->number, "magnifier", 
												  CCMMagnifierOptions[cpt]);
	}
	ccm_screen_plugin_load_options(CCM_SCREEN_PLUGIN_PARENT(plugin), screen);
}

static void
ccm_magnifier_window_load_options(CCMWindowPlugin* plugin, CCMWindow* window)
{
	CCMMagnifier* self = CCM_MAGNIFIER(plugin);
	CCMScreen* screen = ccm_drawable_get_screen(CCM_DRAWABLE(window));
	gint cpt;
	
	for (cpt = 0; cpt < CCM_MAGNIFIER_OPTION_N; cpt++)
	{
		self->priv->options[cpt] = ccm_config_new(screen->number, "magnifier", 
												  CCMMagnifierOptions[cpt]);
	}
	ccm_window_plugin_load_options(CCM_WINDOW_PLUGIN_PARENT(plugin), window);
}

static gboolean
ccm_magnifier_screen_paint(CCMScreenPlugin* plugin, CCMScreen* self)
{
	g_return_if_fail(self != NULL);

	CCMMagnifier* mag = CCM_MAGNIFIER (plugin);
	gboolean ret = FALSE;
	Window win;
	Display *dpy;
	int x, y, w, h;

	ret = ccm_screen_plugin_paint(CCM_SCREEN_PLUGIN_PARENT (self->priv->plugin),
								  self);
	x = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_X]);
	y = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_Y]);
	w = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_WIDTH]);
	h = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_HEIGHT]);

	dpy = CCM_DISPLAY_XDISPLAY (ccm_screen_get_display (self));

	if (!mag->priv->window) {
		win = XCreateSimpleWindow (dpy, CCM_WINDOW_XWINDOW (self->priv->cow),
								   x, y, w, h, 0, 0, 0);
		XMapWindow (dpy, win);
		mag->priv->window = ccm_window_new (self, win);
		ccm_window_make_output_only (mag->priv->window);
		mag->priv->ctx = ccm_drawable_create_context (CCM_DRAWABLE (mag->priv->window));
		cairo_scale (mag->priv->ctx, ccm_config_get_float (mag->priv->options[CCM_MAGNIFIER_ZOOM_LEVEL]), ccm_config_get_float (mag->priv->options[CCM_MAGNIFIER_ZOOM_LEVEL]));
	}

	if (ret)
		ccm_drawable_flush (mag->priv->window);

	return ret;
}

static gboolean
ccm_magnifier_window_paint(CCMWindowPlugin* plugin, CCMWindow* window, 
						   cairo_t* context, cairo_surface_t* surface)
{
	CCMScreen *screen = ccm_drawable_get_screen (CCM_DRAWABLE (window));
	CCMMagnifier* mag = CCM_MAGNIFIER (_ccm_screen_get_plugin (screen, CCM_TYPE_MAGNIFIER));
	gboolean ret;
	Window win;
	Display *dpy;
	int x, y, w, h;

	x = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_X]);
	y = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_Y]);
	w = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_WIDTH]);
	h = ccm_config_get_integer (mag->priv->options [CCM_MAGNIFIER_HEIGHT]);

	dpy = CCM_DISPLAY_XDISPLAY (ccm_screen_get_display (screen));

	if (!mag->priv->window) {
		win = XCreateSimpleWindow (dpy, CCM_WINDOW_XWINDOW (screen->priv->cow),
								   x, y, w, h, 0, 0, 0);
		XMapWindow (dpy, win);
		mag->priv->window = ccm_window_new (screen, win);
		ccm_window_make_output_only (mag->priv->window);
		mag->priv->ctx = ccm_drawable_create_context (CCM_DRAWABLE (mag->priv->window));
		cairo_scale (mag->priv->ctx, ccm_config_get_float (mag->priv->options[CCM_MAGNIFIER_ZOOM_LEVEL]), ccm_config_get_float (mag->priv->options[CCM_MAGNIFIER_ZOOM_LEVEL]));
		cairo_set_source_rgb (mag->priv->ctx, 0, 0, 0);
		cairo_paint(mag->priv->ctx);
	}

	/* paint in the overlay window */
	ret = ccm_window_plugin_paint (CCM_WINDOW_PLUGIN_PARENT(plugin), window, 
								   context, surface);

	/* paint in the magnifier window */
	ret = ccm_window_plugin_paint (CCM_WINDOW_PLUGIN_PARENT(plugin), window,
								   mag->priv->ctx, surface);

	return ret;
}

static void
ccm_magnifier_window_iface_init(CCMWindowPluginClass* iface)
{
	iface->load_options 	= ccm_magnifier_window_load_options;
	iface->query_geometry 	= NULL;
	iface->paint 			= ccm_magnifier_window_paint;
	iface->map				= NULL;
	iface->unmap			= NULL;
	iface->query_opacity  	= NULL;
	iface->set_opaque		= NULL;
	iface->move				= NULL;
	iface->resize			= NULL;
}

static void
ccm_magnifier_screen_iface_init(CCMScreenPluginClass* iface)
{
	iface->load_options 	= ccm_magnifier_screen_load_options;
	iface->paint 			= ccm_magnifier_screen_paint;
	iface->add_window 		= NULL;
	iface->remove_window 	= NULL;
}

